{-# OPTIONS --no-termination-check 
  #-}

module RecursiveNormaliser where
open import Syntax
open import OPE

mutual
  eval : forall {Γ Δ σ} -> Tm Δ σ -> Env Γ Δ -> Val Γ σ
  eval top        (vs << v) = v
  eval (t [ ts ]) vs        = eval t (evalˢ ts vs)
  eval (λ t)      vs        = λv t vs
  eval (t $ u)    vs        = eval t vs $$ eval u vs
  eval void       vs        = voidv
  eval < t , u >  vs        = < eval t vs , eval u vs >v
  eval (fst t)    vs        = vfst (eval t vs) 
  eval (snd t)    vs        = vsnd (eval t vs) 

  vfst : forall {Γ σ τ} -> Val Γ (σ × τ) -> Val Γ σ
  vfst < v , w >v = v
  vfst (nev n)    = nev (fstV n)

  vsnd : forall {Γ σ τ} -> Val Γ (σ × τ) -> Val Γ τ
  vsnd < v , w >v = w
  vsnd (nev n)    = nev (sndV n)

  _$$_ : forall {Γ σ τ} -> Val Γ (σ → τ) -> Val Γ σ -> Val Γ τ
  λv t vs $$ v = eval t (vs << v)
  nev n   $$ v = nev (appV n v)

  evalˢ : forall {Γ Δ Σ} -> Sub Δ Σ -> Env Γ Δ -> Env Γ Σ
  evalˢ (pop σ)   (vs << v) = vs
  evalˢ (ts < t)  vs        = evalˢ ts vs << eval t vs
  evalˢ id        vs        = vs
  evalˢ (ts ○ us) vs        = evalˢ ts (evalˢ us vs)

mutual
  quote : forall {Γ σ} -> Val Γ σ -> Nf Γ σ
  quote {σ = ι}     (nev n) = ne (quoteⁿ n)
  quote {σ = σ → τ} f       = λn (quote (vwk σ f $$ nev (varV vZ)))
  quote {σ = One}   _   = voidn
  quote {σ = σ × τ} p   = < quote (vfst p) , quote (vsnd p) >n   

  -- shouldn't quote return voidn for anything of type One?
  -- but then what happens to neutral terms? Aren't there any?

  quoteⁿ : forall {Γ σ} -> NeV Γ σ -> NeN Γ σ
  quoteⁿ (varV x)   = varN x
  quoteⁿ (appV n v) = appN (quoteⁿ n) (quote v)
  quoteⁿ (fstV n)   = fstN (quoteⁿ n) 
  quoteⁿ (sndV n)   = sndN (quoteⁿ n) 

open import IdentityEnvironment

nf : forall {Γ σ} -> Tm Γ σ -> Nf Γ σ
nf t = quote (eval t vid)
